//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// KBStatistics class - holds the statistics for keyboard usage
//
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ResourceBundle;

/**
 * Holds the statistics for keyboard usage.
 */
public class KBStatistics {

	// a note of the keyboard layout we are monitoring
	private KBLayout	kblayout;//	= null;

	// a hashtable of stats, hashed on character; used to find the
	// glyph statistic to update on a keypress, and the one to
	// return for a requested glyph
	private Hashtable	char_stats_by_char	= new Hashtable();

	// a hashtable of stats, hashed on character; used to find the
	// finger statistic to update on a keypress
	private Hashtable	finger_stats_by_char	= new Hashtable();

	// a hashtable of stats, hashed on finger; used to find the
	// statistic to return for a requested finger
	private Hashtable	finger_stats_by_finger	= new Hashtable();

	// a note of the complete, overall, global statistics
	private KeyingStatistics global_stats;//	= null;


	/**
	 * Constructor to create a customized stats set based on
	 * the keyboard layout passed in.
	 */
	public KBStatistics( KBLayout kblayout ) {

		// note our kblayout
		this.kblayout = kblayout;

		// scan the complete keyboard layout, and build up a data
		// structure that allows us to update details as soon as
		// a key is pressed, and to then retrieve details by char
		// or by finger
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {

			// get the character and finger for this entry
			Glyph glyph		= (Glyph)enum.nextElement();
			Character character	= glyph.get_character();
			Finger finger		= glyph.get_finger();

			// try to get the char statistics for this glyph, and
			// if there aren't any (and there shouldn't be, unless
			// there is a bug in the keyboard layout, and the same
			// glyph is represented more than once), then create
			// the new statistics object for this glyph and place
			// it in the character hash table
			KeyingStatistics char_stats =
					(KeyingStatistics)
					char_stats_by_char.get( character );
			if ( char_stats == null )
				char_stats = new KeyingStatistics();
			char_stats_by_char.put( character, char_stats );

			// if a finger is assigned to operate this glyph,
			// get the finger statistics for this finger, and
			// if there aren't any then create a new set for it,
			// since this must be the first time this finger has
			// been seen, then put the new set into both the
			// by_character and by_finger tables
			if ( finger != Finger.NO_FINGER ) {
				KeyingStatistics finger_stats =
						(KeyingStatistics)
						finger_stats_by_finger.get(
								finger );
				if ( finger_stats == null )
					finger_stats = new KeyingStatistics();
				finger_stats_by_char.put( character,
						finger_stats );
				finger_stats_by_finger.put( finger,
						finger_stats );
			}
		}

		// create space for the global statistics
		global_stats = new KeyingStatistics();
	}


	/**
	 * Update the statistics for a keypress.
	 */
	public void update_statistics( char c, boolean hit, int delta_time ) {

		// create a temporary Character object for hashing
		Character character = new Character( c );

		// find the individual glyph statistics for the character,
		// and if there are any then update them (note that there
		// may not be any if the keyboard layout is incomplete)
		KeyingStatistics glyph_stats =
				(KeyingStatistics)
				char_stats_by_char.get( character );
		if ( glyph_stats != null )
			if ( hit )
				glyph_stats.struck( delta_time );
			else
				glyph_stats.missed( delta_time );

		// now do the exact same thing for the finger statistics
		KeyingStatistics finger_stats =
				(KeyingStatistics)
				finger_stats_by_char.get( character );
		if ( finger_stats != null )
			if ( hit )
				finger_stats.struck( delta_time );
			else
				finger_stats.missed( delta_time );

		// and finally, the global statistics
		if ( hit )
			global_stats.struck( delta_time );
		else
			global_stats.missed( delta_time );
	}

	/**
	 * Find the statistics for a key, a finger, and globally.
	 */
	public KeyingStatistics get_statistics( char character ) {
		return (
		(KeyingStatistics)char_stats_by_char.get(
				new Character( character ) ));
	}
	public KeyingStatistics get_statistics( Character character ) {
		return (
		(KeyingStatistics)char_stats_by_char.get( character ));
	}
	public KeyingStatistics get_statistics( Finger finger ) {
		return (
		(KeyingStatistics)finger_stats_by_finger.get( finger ));
	}
	public KeyingStatistics get_statistics() {
		return ( global_stats );
	}


	/**
	 * Query the KB layout this set is using.
	 */
	public KBLayout get_kblayout() {
		return ( kblayout );
	}


	/**
	 * Reset the complete statistics set to initial state; all the
	 * character statistics, all the finger ones, and the global one
	 * too.
	 */
	public void reset() {
		for ( Enumeration enum = char_stats_by_char.elements();
				enum.hasMoreElements(); ) {
			KeyingStatistics key_stats =
					(KeyingStatistics)enum.nextElement();
			key_stats.reset();
		}
		for ( Enumeration enum = finger_stats_by_finger.elements();
				enum.hasMoreElements(); ) {
			KeyingStatistics key_stats =
					(KeyingStatistics)enum.nextElement();
			key_stats.reset();
		}
		global_stats.reset();
	}


	/**
	 * Printout method.
	 */
	public String toString() {
		return ( getClass().getName() + "{char_stats_by_char@"
				+ char_stats_by_char
				+ ",finger_stats_by_finger@"
				+ finger_stats_by_finger
				+ ",finger_stats_by_char@"
				+ finger_stats_by_char + "}" );
	}


	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		KBLayout k = KBLayout.create_kblayout( "US" );
		KBStatistics ks = new KBStatistics( k );
		Enumeration e = k.get_elements();
		while ( e.hasMoreElements() ) {
			Glyph g = (Glyph)e.nextElement();
			Character key = g.get_character();
			for ( int j = 0; j < key.charValue(); j++ )
				ks.update_statistics( key.charValue(),
						true, j % 100 );
			for ( int j = 0; j < key.charValue() / 10; j++ )
				ks.update_statistics( key.charValue(),
						false, j % 100 );
		}

		System.out.println( ks );
		e = k.get_elements();
		while ( e.hasMoreElements() ) {
			Glyph g = (Glyph)e.nextElement();
			Character key = g.get_character();
			System.out.println( key + " : " +
					ks.get_statistics( key ));
		}

		System.out.println( Finger.LH_THUMB + " : "
				+ ks.get_statistics( Finger.LH_THUMB ));
		System.out.println( Finger.LH_INDEX + " : "
				+ ks.get_statistics( Finger.LH_INDEX ));
		System.out.println( Finger.LH_MIDDLE + " : "
				+ ks.get_statistics( Finger.LH_MIDDLE ));
		System.out.println( Finger.LH_RING + " : "
				+ ks.get_statistics( Finger.LH_RING ));
		System.out.println( Finger.LH_PINKY + " : "
				+ ks.get_statistics( Finger.LH_PINKY ));
		System.out.println( Finger.RH_THUMB + " : "
				+ ks.get_statistics( Finger.RH_THUMB ));
		System.out.println( Finger.RH_INDEX + " : "
				+ ks.get_statistics( Finger.RH_INDEX ));
		System.out.println( Finger.RH_MIDDLE + " : "
				+ ks.get_statistics( Finger.RH_MIDDLE ));
		System.out.println( Finger.RH_RING + " : "
				+ ks.get_statistics( Finger.RH_RING ));
		System.out.println( Finger.RH_PINKY + " : "
				+ ks.get_statistics( Finger.RH_PINKY ));
		System.out.println( "Overall : " + ks.get_statistics() );
	}
}
