//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// StatisticsGUI class - creates a frame to display typing statistics
//
import java.awt.Color;
import java.awt.Frame;
import java.awt.Font;
import java.awt.Panel;
import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.CheckboxMenuItem;
import java.awt.Label;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.MenuShortcut;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;
import java.util.Vector;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.text.MessageFormat;

/**
 * Creates a frame to display typing statistics.
 */
public class StatisticsGUI extends Frame
			implements Runnable,
			ItemListener, ActionListener, WindowListener {
	private boolean inAnApplet = true;

	// message and colour catalog information
	private static final ResourceBundle	messages	=
				ResourceBundle.getBundle( "MessageResource" );
	private static final ResourceBundle	colours		=
				ResourceBundle.getBundle( "ColorResource" );

	// menubar GUI components
	private MenuBar			menu_bar;
	private Menu			file_menu;
	private MenuItem		file_refresh, file_reset, file_close;
	private Menu			view_menu;
	private CheckboxMenuItem	view_finger;//		= null;
	private Vector			view_glyphgroups	= new Vector();
	private Menu			options_menu;
	private CheckboxMenuItem	options_adjWPM;
	private CheckboxMenuItem	options_autorefresh;

	// GUI components
	private Panel		card_panel;
	private CardLayout	card_layout;
	private ImageButton	refresh_button, reset_button;
	private Tooltip		refresh_tooltip, reset_tooltip;
	private Vector		stats_descriptors	= new Vector();

	// number of gauges and labels on the panels we create
	private static final int	COLUMNS_PER_STATISTIC	= 3;
	private static final int	STATISTICS_PER_ROW	= 3;

	// somewhat arbitrary limits for warnings/errors on gauges; this
	// does the coloured gauges to try to help pick out the keys that
	// need more practice
	private static final int	ERROR_ACCURACY	=
			Integer.parseInt( messages.getString(
					"StatisticsGUI.erroraccuracy" ));
	private static final int	WARN_ACCURACY	=
			Integer.parseInt( messages.getString(
					"StatisticsGUI.warnaccuracy" ));
	private static final int	ERROR_SPEED	=
			Integer.parseInt( messages.getString(
					"StatisticsGUI.errorspeed" ));
	private static final int	WARN_SPEED	=
			Integer.parseInt( messages.getString(
					"StatisticsGUI.warnspeed" ));

	// standard colours for various conditions
	private static final Color	COLOUR_GAUGEBG	=
			Colour.find_colour( colours.getString(
					"StatisticsGUI.gaugebg" ));
	private static final Color	COLOUR_INACTIVE_GAUGEBG	=
			Colour.find_colour( colours.getString(
					"StatisticsGUI.inactivegaugebg" ));
	private static final Color	COLOUR_GAUGETXT	=
			Colour.find_colour( colours.getString(
					"StatisticsGUI.gaugetext" ));
	private static final Color	COLOUR_GOOD	=
			Colour.find_colour( colours.getString(
					"StatisticsGUI.gaugegood" ));
	private static final Color	COLOUR_FAIR	=
			Colour.find_colour( colours.getString(
					"StatisticsGUI.gaugefair" ));
	private static final Color	COLOUR_BAD	=
			Colour.find_colour( colours.getString(
					"StatisticsGUI.gaugebad" ));

	// statistics set we are working with, for reference
	private KBStatistics	stats_set;//	= null;

	// synchronization monitor, for setVisible notifications
	private Object		visible		= new Object();


	// inner class tying together a set of statistics and the gauges
	// that represent it
	class StatsDescriptor {
		protected KeyingStatistics	key_stats;//		= null;
		protected Gauge			accuracy_gauge;//	= null;
		protected Gauge			speed_gauge;//		= null;
		protected StatsDescriptor( KeyingStatistics key_stats,
				Gauge accuracy_gauge, Gauge speed_gauge ) {
			this.key_stats		= key_stats;
			this.accuracy_gauge	= accuracy_gauge;
			this.speed_gauge	= speed_gauge;
		}
	}

	/**
	 * Initialize the whole GUI set.
	 */
	public StatisticsGUI( boolean inAnApplet, KBStatistics stats_set ) {
		this.addWindowListener( this );
		this.inAnApplet = inAnApplet;

		// create GUI components
		setTitle( messages.getString( "StatisticsGUI.title" ));
		setLayout( new BorderLayout() );
		setForeground( Colour.find_colour(
			colours.getString( "StatisticsGUI.foreground" )));
		setBackground( Colour.find_colour(
			colours.getString( "StatisticsGUI.background" )));
		setFont( Font.decode(
			colours.getString( "StatisticsGUI.font" )));

		// note the stats set for reference
		this.stats_set = stats_set;

		// create the menu bar stuff
		menu_bar = new MenuBar();
		setMenuBar( menu_bar );
		// File menu
		file_menu = new Menu(
			messages.getString( "StatisticsGUI.file" ));
		menu_bar.add( file_menu );
		file_refresh = new MenuItem(
			messages.getString( "StatisticsGUI.file_refresh" ),
			new MenuShortcut(
				messages.getString(
				"StatisticsGUI.file_refresh_sc" ).charAt(0) ));
		file_menu.add( file_refresh );
		file_reset = new MenuItem(
			messages.getString( "StatisticsGUI.file_reset" ),
			new MenuShortcut(
				messages.getString(
				"StatisticsGUI.file_reset_sc" ).charAt(0) ));
		file_reset.addActionListener( this );
		file_menu.add( file_reset );
		file_close = new MenuItem(
			messages.getString( "StatisticsGUI.file_close" ),
			new MenuShortcut(
				messages.getString(
				"StatisticsGUI.file_close_sc" ).charAt(0) ));
		file_close.addActionListener( this );
		file_menu.add( file_close );

		// first part of the view menu
		view_menu = new Menu(
			messages.getString( "StatisticsGUI.view" ));
		menu_bar.add( view_menu );

		// create the refresh and reset buttons
		Panel toolbar = new Panel();
		toolbar.setLayout( new FlowLayout( FlowLayout.LEFT ));
		add( "North", toolbar );

		refresh_button = new ImageButton(
			Xpm.get_image_producer(
				messages.getString( "StatisticsGUI.refreshxpm" )),
			messages.getString( "StatisticsGUI.refreshxpmlabel" ));
		refresh_tooltip = new Tooltip( this, refresh_button,
			messages.getString( "StatisticsGUI.refreshtooltip" ));
		refresh_button.addActionListener( this );
		toolbar.add( refresh_button );
		reset_button = new ImageButton(
			Xpm.get_image_producer(
				messages.getString( "StatisticsGUI.resetxpm" )),
			messages.getString( "StatisticsGUI.resetxpmlabel" ));
		reset_tooltip = new Tooltip( this, reset_button,
			messages.getString( "StatisticsGUI.resettooltip" ));
		reset_button.addActionListener( this );
		toolbar.add( reset_button );

		String[] fmtargs = { stats_set.get_kblayout().get_kb_id() };
		Label kbtype_label = new Label(
			MessageFormat.format(
				messages.getString( "StatisticsGUI.kbtype" ),
				fmtargs ), Label.RIGHT );
		add( "South", kbtype_label );

		// create a card panel for the stats to live on
		card_panel	= new Panel();
		card_layout	= new CardLayout();
		card_panel.setLayout( card_layout );

		// find the maximum number of statistics any panel
		// in the set has; this is used for padding the panels
		// with null labels so the sizes of all the gauges and
		// labels are equal
		int max_num_stats = -1;
		for ( Enumeration enum = GlyphGroup.get_elements();
				enum.hasMoreElements(); ) {
			GlyphGroup glyphgroup = (GlyphGroup)enum.nextElement();

			// see if this group has more statistics
			max_num_stats = Math.max(
					size_group_panel( stats_set,
							glyphgroup ),
					max_num_stats );
		}
		max_num_stats = Math.max(
				size_finger_panel( stats_set ),
				max_num_stats );

		// from this number, calculate the number of statistics,
		// real or padded, we want to put on each and every panel
		int stats_per_panel = ( (max_num_stats - 1) /
						STATISTICS_PER_ROW )
					* STATISTICS_PER_ROW
					+ STATISTICS_PER_ROW;

		// create the panel of finger statistics, and a menu item
		Panel fstats_panel = make_finger_stats_panel( stats_set,
				stats_per_panel );
		fstats_panel.setForeground( Colour.find_colour(
			colours.getString( "StatisticsGUI.panel_foreground" )));
		fstats_panel.setBackground( Colour.find_colour(
			colours.getString( "StatisticsGUI.panel_background" )));
		fstats_panel.setFont( Font.decode(
			colours.getString( "StatisticsGUI.panel_font" )));
		view_finger = new CheckboxMenuItem(
			messages.getString( "StatisticsGUI.view_finger" ));
		view_finger.addItemListener( this );
		view_menu.add( view_finger );
		card_panel.add( view_finger.getLabel(), fstats_panel );

		// create a panel for each of the glyph groups defined,
		// and add a view menuitem so that each can be accessed
		for ( Enumeration enum = GlyphGroup.get_elements();
				enum.hasMoreElements(); ) {
			GlyphGroup glyphgroup = (GlyphGroup)enum.nextElement();

			// create a panel for this group
			Panel stats_panel = make_stats_panel( stats_set,
					glyphgroup, stats_per_panel );

			// see if there is anything on the panel
			if ( stats_panel != null ) {

				// get a string to identify this glyph group
				String glyphgroup_text =
					messages.getString(
						glyphgroup.get_resource_id() );

				// create a view menu item for this group, and
				// add it to the list we use when panels change
				CheckboxMenuItem cbmenuitem
						= new CheckboxMenuItem(
							glyphgroup_text );
				cbmenuitem.addItemListener( this );
				view_menu.add( cbmenuitem );
				view_glyphgroups.addElement( cbmenuitem );

				// set the panel background, and add it to the
				// card layout, with the keystring equal to
				// the text in the menu item -- this makes
				// it very easy to find from the menu item
				stats_panel.setForeground(
					Colour.find_colour( colours.getString(
					"StatisticsGUI.panel_foreground" )));
				stats_panel.setBackground(
					Colour.find_colour( colours.getString(
					"StatisticsGUI.panel_background" )));
				card_panel.add( cbmenuitem.getLabel(),
						stats_panel );
			}
		}

		// options menu
		options_menu = new Menu(
			messages.getString( "StatisticsGUI.options" ));
		menu_bar.add( options_menu );
		options_adjWPM = new CheckboxMenuItem(
			messages.getString( "StatisticsGUI.options_adjwpm" ));
		options_adjWPM.addItemListener( this );
		options_menu.add( options_adjWPM );
		options_autorefresh = new CheckboxMenuItem(
			messages.getString( "StatisticsGUI.options_autorefresh" ));
		options_menu.add( options_autorefresh );

		// add the card panel to the main panel
		view_finger.setState( true );
		card_layout.show( card_panel, view_finger.getLabel() );
		add( "Center", card_panel );

		// start the refresher thread
		new Thread( this, "StatisticsGUI_refresh" ).start();
	}


	/**
	 * Thread to automatically call refresh, if the option to do it
	 * is set.
	 */
	public void run() {

		// thread 'tick', in mS
		final int	THREAD_TICK	=
				Integer.parseInt( messages.getString(
					"StatisticsGUI.autorefreshperiod" ));

		// loop forever, at the lowest priority
		Thread.currentThread().setPriority( Thread.MIN_PRIORITY );
		while ( true ) {

			// if our window isn't visible then just do nothing
			while ( ! isVisible() ) {
				try { synchronized( visible ) {
					visible.wait();
				}}
				catch ( InterruptedException e ) {}
			}

			// wait until the next tick
			try { Thread.sleep( THREAD_TICK ); }
			catch ( InterruptedException e ) {}

			// if the option is set, call refresh
			if ( options_autorefresh.getState() )
				refresh();
		}
	}


	/*
	 * Count how many statistics we would display on a panel, given
	 * this set and grouping - used to search for largest group.
	 */
	private int size_group_panel( KBStatistics stats_set,
			GlyphGroup group ) {

		// find the KB layout for the stats
		KBLayout kblayout = stats_set.get_kblayout();

		// now count the statistics for this group
		int num_stats = 0;
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {
			Glyph glyph = (Glyph)enum.nextElement();

			// see if this matches the target group
			if ( glyph.get_group() == group )

				// increment the count if there are stats
				if ( stats_set.get_statistics(
						glyph.get_character() )
						!= null )
					num_stats++;
		}

		// return the count
		return ( num_stats );
	}


	/*
	 * Do the same sizing exercise for the finger panel.
	 */
	private int size_finger_panel( KBStatistics stats_set ) {

		// now count the statistics for fingers represented
		// in this stats set, starting at one because this panel
		// will also contain the global statistics
		int num_stats = 1;
		for ( Enumeration enum = Finger.get_elements();
				enum.hasMoreElements(); ) {
			Finger finger = (Finger)enum.nextElement();

			// increment the count if there are stats
			if ( stats_set.get_statistics( finger ) != null )
				num_stats++;
		}

		// return the count
		return ( num_stats );
	}


	/*
	 * Create a panel containing finger statistics.
	 */
	private Panel make_finger_stats_panel( KBStatistics stats_set,
			int stats_per_panel ) {

		// start with a set of column headings
		Panel panel = new Panel();
		int columns = COLUMNS_PER_STATISTIC * STATISTICS_PER_ROW;
		panel.setLayout( new GridLayout( 0, columns ));
		for ( int index = 0; index < STATISTICS_PER_ROW; index++ ) {
			panel.add( new Label() );
			panel.add( new Label(
				messages.getString( "StatisticsGUI.accuracy" ),
					Label.CENTER ));
			panel.add( new Label(
				messages.getString( "StatisticsGUI.speed" ),
					Label.CENTER ));
		}

		// add data labels for all the fingers we could monitor;
		// note that if a finger actually does no keys then it
		// will not appear on the stats panel
		int num_stats = 0;
		KeyingStatistics key_stats = null;
		for ( Enumeration enum = Finger.get_elements();
				enum.hasMoreElements(); ) {
			Finger finger = (Finger)enum.nextElement();

			// get finger statistics
			key_stats = stats_set.get_statistics( finger );

			// if there are any then add to the panel
			if ( key_stats != null ) {
				make_disp_entry( panel, key_stats,
					messages.getString(
						finger.get_resource_id() ));
				num_stats++;
			}
		}

		// pad the panel to the maximum size, less one for the
		// global statistics
		for ( ; num_stats < stats_per_panel - 1; num_stats ++ )
			make_disp_entry( panel, null, null );

		// add on, at the very end, the global statistics
		key_stats = stats_set.get_statistics();
		make_disp_entry( panel, key_stats,
			messages.getString( "StatisticsGUI.overall" ));

		// return the panel - note how this one can't be null since
		// there must be some statistics on it!
		return ( panel );
	}


	/*
	 * Create a complete panel of labels/gauges given a char group, keymap
	 * table, and some luck.
	 */
	private Panel make_stats_panel( KBStatistics stats_set,
			GlyphGroup char_group, int stats_per_panel ) {

		// create the new panel and add headings
		Panel panel = new Panel();
		int columns = COLUMNS_PER_STATISTIC * STATISTICS_PER_ROW;
		panel.setLayout( new GridLayout( 0, columns ));
		for ( int index = 0; index < STATISTICS_PER_ROW; index++ ) {
			panel.add( new Label() );
			panel.add( new Label(
				messages.getString( "StatisticsGUI.accuracy" ),
					Label.CENTER ));
			panel.add( new Label(
				messages.getString( "StatisticsGUI.speed" ),
					Label.CENTER ));
		}

		// find the KB layout for the stats
		KBLayout kblayout = stats_set.get_kblayout();

		// now add the data gauges
		int num_stats = 0;
		for ( Enumeration enum = kblayout.get_elements();
				enum.hasMoreElements(); ) {
			Glyph glyph = (Glyph)enum.nextElement();

			// if this char is in our group then process it
			if ( glyph.get_group() == char_group ) {

				// figure out a legend for this entry
				Character character = glyph.get_character();
				String legend = null;
				if ( character.charValue() == ' ' )
					legend = messages.getString(
						"StatisticsGUI.space" );
				else if ( character.charValue() == '\n' )
					legend = messages.getString(
						"StatisticsGUI.return" );
				else {
					String[] fmtargs = { "" + character };
					legend = MessageFormat.format(
						messages.getString(
						"StatisticsGUI.key" ),
						fmtargs );
				}

				// create the stats entry for this key
				make_disp_entry( panel,
					stats_set.get_statistics( character ),
					legend );
				num_stats++;
			}
		}

		// return null if nothing on the panel, otherwise pad the
		// panel to the largest size and return it
		if ( num_stats > 0 ) {

			// pad the panel to the maximum size
			for ( ; num_stats < stats_per_panel; num_stats++ )
				make_disp_entry( panel, null, null );

			// return the panel
			return ( panel );
		}
		else
			// nothing on this one - return null
			return ( null );
	}


	/*
	 * Create a trio of components showing the specified stats
	 * set and add the trio to the panel given.
	 */
	private void make_disp_entry( Panel panel, KeyingStatistics key_stats,
			String legend ) {

		// if there is a statistics source then add the gauges
		if ( key_stats != null && legend != null ) {

			// create new gauges for this statistic
			Gauge	accuracy_gauge	= new Gauge( 0, 100, 0,
				messages.getString( "StatisticsGUI.pc" ),
				messages.getString( "StatisticsGUI.nopcdata" ));
			Gauge	speed_gauge	= new Gauge( 0, 100, 0,
				messages.getString( "StatisticsGUI.wpm" ),
				messages.getString( "StatisticsGUI.nowpmdata" ));

			// add a legend label and the two readout gauges
			panel.add( new Label( legend, Label.RIGHT ));
			accuracy_gauge.setBackground( COLOUR_GAUGEBG );
			accuracy_gauge.setForeground( COLOUR_GAUGETXT );
			accuracy_gauge.set_colour( COLOUR_GOOD );
			accuracy_gauge.set_inactive_bgcolour(
						COLOUR_INACTIVE_GAUGEBG );
			panel.add( accuracy_gauge );
			speed_gauge.setBackground( COLOUR_GAUGEBG );
			speed_gauge.setForeground( COLOUR_GAUGETXT );
			speed_gauge.set_colour( COLOUR_GOOD );
			speed_gauge.set_inactive_bgcolour(
						COLOUR_INACTIVE_GAUGEBG );
			panel.add( speed_gauge );

			// create a descriptor for this statistic
			StatsDescriptor descriptor = new StatsDescriptor(
					key_stats,
					accuracy_gauge, speed_gauge );
			stats_descriptors.addElement( descriptor );
		}
		else {
			// add a null entry for padding - three blank labels
			panel.add( new Label() );
			panel.add( new Label() );
			panel.add( new Label() );
		}
	}


	/**
	 * Refresh the gauges with the latest stats data.
	 */
	public void refresh() {

		// note if we are doing adjusted timings - this gets the
		// option getState out of the loop
		boolean adjusted = options_adjWPM.getState();

		// run through the list of statistics we need to update
		for ( int index = 0; index < stats_descriptors.size();
					index++ ) {

			// get the keying stats we have to use, and
			// the gauges we are going to update the
			// data into
			StatsDescriptor descriptor = (StatsDescriptor)
					stats_descriptors.elementAt( index );
			KeyingStatistics
				key_stats	= descriptor.key_stats;
			Gauge	accuracy_gauge	= descriptor.accuracy_gauge;
			Gauge	speed_gauge	= descriptor.speed_gauge;

			// see if this statistic has anything to tell us
			if ( key_stats.has_statistics() ) {

				// make the gauges active
				accuracy_gauge.set_active( true );
				speed_gauge.set_active( true );

				// get the statistics values
				int accuracy	= key_stats.get_pc_accuracy();
				int speed	= key_stats.get_WPM( adjusted );

				// update the gauges with current values
				accuracy_gauge.set_value( accuracy );
				if ( accuracy <= ERROR_ACCURACY )
					accuracy_gauge.set_colour(
								COLOUR_BAD );
				else if ( accuracy <= WARN_ACCURACY )
					accuracy_gauge.set_colour(
								COLOUR_FAIR );
				else
					accuracy_gauge.set_colour(
								COLOUR_GOOD );
				speed_gauge.set_value( speed );
				if ( speed <= ERROR_SPEED )
					speed_gauge.set_colour(
								COLOUR_BAD );
				else if ( speed <= WARN_SPEED )
					speed_gauge.set_colour(
								COLOUR_FAIR );
				else
					speed_gauge.set_colour(
								COLOUR_GOOD );
			}
			else {

				// reset the gauges to nothing
				accuracy_gauge.set_active( false );
				speed_gauge.set_active( false );
			}
		}
	}


	/**
	 * Override setVisible so that the statistics get refreshed before
	 * the window is drawn, and the thread is kicked.
	 */
	public void setVisible( boolean is_visible ) {
		super.setVisible( is_visible );
		if ( is_visible ) {
			synchronized( visible ) { visible.notify(); }

			// refresh the complete stats set
			refresh();
		}
	}

	/*
	 * Handle dismiss, close, etc by just hiding ourselves.
	 */
	private void handle_close_button() {

		// simply just hide this frame from view
		setVisible( false );
	}


	/*
	 * Handle reset request by resetting the underlying stats
	 * then redisplaying.
	 */
	private void handle_reset_button() {
		stats_set.reset();
		refresh();
	}


	/*
	 * Handle view change by displaying the card selected.
	 */
	private void handle_view_change( CheckboxMenuItem selected ) {

		// emulate the groups thing by deselecting all menu items,
		// then reselecting just the selected one
		view_finger.setState( false );
		for ( Enumeration enum = view_glyphgroups.elements();
				enum.hasMoreElements(); ) {
			CheckboxMenuItem cbmenuitem = (CheckboxMenuItem)
					enum.nextElement();

			// unset this glyphgroup view menu item
			cbmenuitem.setState( false );
		}
		selected.setState( true );

		// display the right card panel, using the label string
		// as the key to the card
		card_layout.show( card_panel, selected.getLabel() );
		refresh();
	}

	/*
	 * Handle GUI component actions and events.
	 */
	public void actionPerformed( ActionEvent evt ) {
		if ( evt.getSource() instanceof ImageButton ) {
			if ( evt.getSource() == refresh_button ) {
				refresh();
			}
			else if ( evt.getSource() == reset_button ) {
				handle_reset_button();
			}
		}
		else if ( evt.getSource() instanceof MenuItem ) {
			if ( evt.getSource() == file_close ) {
				handle_close_button();
			}
			else if ( evt.getSource() == file_refresh ) {
				refresh();
			}
			else if ( evt.getSource() == file_reset ) {
				handle_reset_button();
			}
		}
	}
	public void windowClosing( WindowEvent evt ) {
		handle_close_button();
	}
	public void windowClosed( WindowEvent evt ) {}
	public void windowOpened( WindowEvent evt ) {
		refresh_button.requestFocus();
	}
	public void windowIconified( WindowEvent evt ) {}
	public void windowDeiconified( WindowEvent evt ) {
		refresh_button.requestFocus();
	}
	public void windowActivated( WindowEvent evt ) {
		refresh_button.requestFocus();
	}
	public void windowDeactivated( WindowEvent evt ) {}
	public void itemStateChanged( ItemEvent evt ) {
		if ( evt.getSource() instanceof CheckboxMenuItem ) {
			if ( evt.getSource() == options_adjWPM )
				refresh();
			else if ( evt.getSource() != options_autorefresh )
				handle_view_change(
					(CheckboxMenuItem)evt.getSource() );
		}
	}


	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		KBStatistics kk = new KBStatistics(
				KBLayout.create_kblayout( "US" ) );
		kk.update_statistics( ' ', true, 200 );
		StatisticsGUI stats = new StatisticsGUI( false, kk );
		stats.pack(); stats.setVisible( true );
		kk.update_statistics( 'q', true, 200 );
		kk.update_statistics( 'q', true, 200 );
		kk.update_statistics( 'q', false, 50 );
	}
}
